/* $Id: ldpc_fec.cpp,v 1.77 2006/03/09 08:35:13 roca Exp $ */
/*
 *  LDPC/LDGM FEC Library.
 *  (c) Copyright 2002-2006 INRIA - All rights reserved
 *  Main authors: Christoph Neumann (christoph.neumann@inrialpes.fr)
 *                Vincent Roca      (vincent.roca@inrialpes.fr)
 *		  Julien Laboure   (julien.laboure@inrialpes.fr)
 *
 *  This copyright notice must be retained and prominently displayed,
 *  along with a note saying that the original programs are available from
 *  Vincent Roca's web page, and note is made of any changes made to these
 *  programs.  
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ldpc_fec.h"
#include "macros.h"
#ifdef WIN32
#include <conio.h>
#endif


/******************************************************************************
 * LDPCFecSession Contructor.
 */
LDPCFecSession::LDPCFecSession()
{
	memset(this, 0, sizeof(*this));
}


/******************************************************************************
 * LDPCFecSession Destructor.
 */
LDPCFecSession::~LDPCFecSession()
{
        EndSession();
}


/******************************************************************************
 * InitSession : Initializes the LDPC session.
 * => See header file for more informations.
 */
ldpc_error_status
LDPCFecSession::InitSession (	int nbSourceSymbols,
				int nbParitySymbols,
				int symbolSize,
				int flags,
				int seed,
				SessionType codecType,
				int leftDegree)
{
	mod2entry	*e;

	m_initialized	= false;
	m_sessionFlags	= flags;
	m_sessionType	= codecType;
	m_symbolSize	= symbolSize;

#ifdef PERF_COUNT_XOR
	m_nbXor = 0;
#endif 
 
	if ((symbolSize % 4) != 0) {
		fprintf(stderr, "LDPCFecSession::InitSession: ERROR: Symbol size (%d bytes) IS NOT multiple of 4\n", symbolSize);
		return LDPC_ERROR;
	}
#if defined (__LP64__) || (__WORDSIZE == 64)
	// symbolSize is not necessarily a multiple of 8, but >> 3 will divide
	// it by 8 and keep the integral part automatically.
	m_symbolSize64	= symbolSize >> 3;
#endif 
	m_symbolSize32	= symbolSize >> 2;

	m_nbSourceSymbols	= nbSourceSymbols;
	m_nbParitySymbols	= nbParitySymbols;
	// sanity check: setting a left degree != 3 is only meaningfull with
	// LDGM codes, not with LDPC-Staircase/Triangle codes.
	if ((m_sessionType != TypeLDGM) && (leftDegree != 3)) {
		fprintf(stderr, "LDPCFecSession::InitSession: ERROR: setting a leftDegree!=3 (here %d) is only meaningfull with LDGM codes, not with LDPC-Staircase/Triangle codes!\n", leftDegree);
		return LDPC_ERROR;
	}
	m_leftDegree	= leftDegree;
	m_firstNonDecoded = 0;

	if (this->m_verbosity >= 1) {
		if (m_sessionType == TypeLDGM) {
			printf("Initializing LDGM FEC Session...\n - Source symbols = %d\n - Parity symbols = %d\n - Symbol size = %d\n - Edges per source symbols = %d\n", m_nbSourceSymbols, m_nbCheck, m_symbolSize, m_leftDegree = leftDegree);
		} else if (m_sessionType == TypeSTAIRS) {
			printf("Initializing LDGM STAIRCASE FEC Session...\n - Source symbols = %d\n - Parity symbols = %d\n - Symbol size = %d\n - Edges per source symbols = %d\n", m_nbSourceSymbols, m_nbCheck, m_symbolSize, m_leftDegree = leftDegree);
		} else if (m_sessionType == TypeTRIANGLE) {
			printf("Initializing LDGM TRIANGLE FEC Session...\n - Source symbols = %d\n - Parity symbols = %d\n - Symbol size = %d\n - Edges per source symbols = %d\n", m_nbSourceSymbols, m_nbCheck, m_symbolSize, m_leftDegree = leftDegree);
		} 
	}

	// generate parity check matrix... 
	if (this->m_verbosity >= 1) {
		printf("Generating Parity Check Matrix (H)...");
	}
	m_pchkMatrix = CreatePchkMatrix(m_nbCheck, m_nbSourceSymbols + m_nbParitySymbols, Evenboth, m_leftDegree, seed, false, m_sessionType, this->m_verbosity);
	if (m_pchkMatrix == NULL) {
		fprintf(stderr, "LDPCFecSession::InitSession: ERROR: call to CreatePchkMatrix failed!\n");
		return LDPC_ERROR;
	}
	if (this->m_verbosity >= 1) {
		printf("Done!\n");
	}

	if (m_sessionFlags & FLAG_CODER) {
		m_nb_unknown_symbols_encoder = (int*)calloc(m_nbCheck, sizeof(int));
		if (m_nb_unknown_symbols_encoder == NULL) {
			fprintf(stderr, "LDPCFecSession::InitSession: ERROR: call to calloc failed for m_nb_unknown_symbols_encoder!\n");
			return LDPC_ERROR;
		}

		for (int row=0; row<m_nbCheck; row++) {
			mod2entry *e;
			for (e = mod2sparse_first_in_row(m_pchkMatrix, row);
			     !mod2sparse_at_end(e);
			     e = mod2sparse_next_in_row(e))
			{
				m_nb_unknown_symbols_encoder[row]++;
			}
		}
	} else {
		m_nb_unknown_symbols_encoder = NULL;
	}

	if (m_sessionFlags & FLAG_DECODER) {
		// allocate all internal tables
		if (((m_checkValues	= (void**)calloc(m_nbCheck, sizeof(void*))) == NULL) ||
		    ((m_nbSymbols_in_equ = (int*)calloc(m_nbCheck, sizeof(int))) == NULL) ||
		    ((m_nb_unknown_symbols = (int*)calloc(m_nbCheck, sizeof(int))) == NULL) ||
		    ((m_nbEqu_for_parity = (int*)calloc(m_nbCheck, sizeof(int))) == NULL) ||
		    ((m_parity_symbol_canvas = (void**)calloc(m_nbCheck, sizeof(void*))) == NULL)) {
			fprintf(stderr, "LDPCFecSession::InitSession: ERROR: call to calloc failed for m_parity_symbol_canvas!\n");
			return LDPC_ERROR;
		}
		// and update the various tables now
		for (int row = 0; row < m_nbCheck; row++) {
			for (e = mod2sparse_first_in_row(m_pchkMatrix, row);
			     !mod2sparse_at_end(e);
			     e = mod2sparse_next_in_row(e))
			{
				m_nbSymbols_in_equ[row]++;
				m_nb_unknown_symbols[row]++;
			}
		}
		for (int seq = m_nbSourceSymbols; seq < m_nbTotalSymbols; seq++) {
			for (e = mod2sparse_first_in_col(m_pchkMatrix,
						    GetMatrixCol(seq));
			     !mod2sparse_at_end(e);
			     e = mod2sparse_next_in_col(e))
			{
				m_nbEqu_for_parity[seq - m_nbSourceSymbols]++;
			}
		}
	} else {
		// CODER session
		m_checkValues = NULL;
		m_nbSymbols_in_equ = NULL;
		m_nb_unknown_symbols = NULL;
		m_nbEqu_for_parity = NULL;
		m_parity_symbol_canvas = NULL;
	}
	if ((m_sessionType == TypeTRIANGLE) && ((m_nbTotalSymbols/m_nbSourceSymbols) < 2.0)) {
		m_triangleWithSmallFECRatio = true;
	} else {
		m_triangleWithSmallFECRatio = false;
	}
#ifdef DEBUG
	if (this->m_verbosity >= 2) {
		printf("Pchk Matrix:\n");
		mod2sparse_print(stdout, m_pchkMatrix);
	}
#endif
	m_initialized = true;
	//printf("Pchk Matrix:\n");
	//mod2sparse_print(stdout, m_pchkMatrix);
	return LDPC_OK;
}


/******************************************************************************
 * SetDecodedFunctions: Call the function whenever BuildParitySymbol decodes
 * a new parity or source symbol.
 * => See header file for more informations.
 */
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
ldpc_error_status
LDPCFecSession::SetCallbackFunctions (
		void* (*DecodedSymbol_callback)		(void *context, int size,
							 int symbol_seqno),
		void* (*AllocTmpBuffer_callback)	(void *context, int size),
		void* (*GetData_callback)		(void *context, void *symbol),
		void* (*GetDataPtrOnly_callback)	(void *context, void *symbol),
		ldpc_error_status (*StoreData_callback) (void *context, void *symbol),
		ldpc_error_status (*FreeSymbol_callback)	(void *context, void *symbol),
		void*	context_4_callback)
{
	// sanity checks first
	if (DecodedSymbol_callback) {
		if (!(m_sessionFlags & FLAG_DECODER)) {
			fprintf(stderr, "LDPCFecSession::SetCallbackFunctions: ERROR: specifying DecodedSymbol_callback is only valid in DECODER mode\n");
			return LDPC_ERROR;
		}
	}
	if (AllocTmpBuffer_callback) {
		if (!(m_sessionFlags & FLAG_DECODER)) {
			fprintf(stderr, "LDPCFecSession::SetCallbackFunctions: ERROR: specifying AllocTmpBuffer_callback is only valid in DECODER mode\n");
			return LDPC_ERROR;
		}
	}
	if (FreeSymbol_callback) {
		if (!(m_sessionFlags & FLAG_DECODER)) {
			fprintf(stderr, "LDPCFecSession::SetCallbackFunctions: ERROR: specifying FreeSymbol_callback is only valid in DECODER mode\n");
			return LDPC_ERROR;
		}
	}
	// then remember everything
	m_decodedSymbol_callback = DecodedSymbol_callback;
	m_allocTmpBuffer_callback = AllocTmpBuffer_callback;
	m_getData_callback = GetData_callback;
	m_getDataPtrOnly_callback = GetDataPtrOnly_callback;
	m_storeData_callback = StoreData_callback;
	m_freeSymbol_callback = FreeSymbol_callback;
	m_context_4_callback = context_4_callback;
	return LDPC_OK;
}

#else  /* !EXTERNAL_MEMORY_MGMT_SUPPORT */

ldpc_error_status
LDPCFecSession::SetCallbackFunctions (
		void* (*DecodedSymbol_callback)	(void *context, int size, int symbol_seqno),
		void*	context_4_callback)
{
	// sanity checks first
	if (DecodedSymbol_callback) {
		if (!(m_sessionFlags & FLAG_DECODER)) {
			fprintf(stderr, "LDPCFecSession::SetCallbackFunctions: ERROR: specifying DecodedSymbol_callback is only valid in DECODER mode\n");
			return LDPC_ERROR;
		}
	}
	// then remember everything
	m_decodedSymbol_callback = DecodedSymbol_callback;
	m_context_4_callback = context_4_callback;
	return LDPC_OK;
}
#endif /* !EXTERNAL_MEMORY_MGMT_SUPPORT */


/******************************************************************************
 * EndSession : Ends the LDPC session, and cleans up everything.
 * => See header file for more informations.
 */
void
LDPCFecSession::EndSession()
{
	if (m_initialized == true) {
		m_initialized = false;
		mod2sparse_free(m_pchkMatrix);
		free(m_pchkMatrix);	/* mod2sparse_free does not free it! */

		if (m_checkValues != NULL) {
			for (int i = 0; i < m_nbCheck; i++) {
				if (m_checkValues[i] != NULL) {
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
					if (m_freeSymbol_callback != NULL) {
						m_freeSymbol_callback(m_context_4_callback,
								   m_checkValues[i]);
					} else
#endif
					{
						free(m_checkValues[i]);
					}
				}
			}
			free(m_checkValues);
		}
		if (m_parity_symbol_canvas != NULL) {
			for (int i = 0; i < m_nbCheck; i++) {
				if (m_parity_symbol_canvas[i] != NULL) {
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
					if (m_freeSymbol_callback != NULL) {
						m_freeSymbol_callback(m_context_4_callback,
								   m_parity_symbol_canvas[i]);
					} else
#endif
					{
						free(m_parity_symbol_canvas[i]);
					}
				}
			}
			free(m_parity_symbol_canvas);
		}
		if (m_nbSymbols_in_equ != NULL) {
			free(m_nbSymbols_in_equ);
		}
		if (m_nbEqu_for_parity != NULL) {
			free(m_nbEqu_for_parity);
		}
		if (m_nb_unknown_symbols != NULL) {
			free(m_nb_unknown_symbols);
		}
		if (m_nb_unknown_symbols_encoder != NULL) {
			free(m_nb_unknown_symbols_encoder);
		}
	}
	// and now init everything!
	memset(this, 0, sizeof(*this));
}


/******************************************************************************
 * SetVerbosity: Sets the verbosity level.
 * => See header file for more informations.
 */
void
LDPCFecSession::SetVerbosity(int	verb)
{
	this->m_verbosity = verb;
}


/******************************************************************************
 * MoreAbout:Prints version number and copyright information about this codec.
 * => See header file for more informations.
 */
void
LDPCFecSession::MoreAbout (FILE		*out)
{
	fprintf(out, "LDPC/LDGM large block FEC codec - Version 2.0, March 9th, 2006\n");
	fprintf(out, "  Copyright (c) 2002-2006 INRIA - All rights reserved\n");
	fprintf(out, "  Authors: C. Neumann, V. Roca, L. Fazio, J. Laboure\n");
	fprintf(out, "  This codec contains code from R. Neal:\n");
	fprintf(out, "  Copyright (c) 1995-2003 by Radford M. Neal\n");
	fprintf(out, "  See the associated LICENCE.TXT file for licence information\n");
	switch (m_sessionType) {
	case TypeLDGM:
		fprintf(out, "  LDGM codec\n");
		break;
	case TypeSTAIRS:
		fprintf(out, "  LDPC-Staircase codec\n");
		break;
	case TypeTRIANGLE:
		fprintf(out, "  LDPC-Triangle codec\n");
		break;
	}
}


/******************************************************************************
 * Calculates the XOR sum of two symbols: to = to + from.
 * => See header file for more informations.
 */
void
LDPCFecSession::AddToSymbol	(void	*to,
				void	*from)
{
	uintptr_t	offset;			
#if defined (__LP64__) || (__WORDSIZE == 64)
	// 64-bit machines
	for (offset = 0; offset < m_symbolSize64; offset++) {
		*(((uintptr_t*)to) + offset) ^= *(((uintptr_t*)from) + offset);
#ifdef PERF_COUNT_XOR
		m_nbXor++;
#endif
	}
	/* add the last 32 bits if needed */
	if ((m_symbolSize64 << 1) < m_symbolSize32) {
		*(uint32_t*)(((uintptr_t*)to) + offset) ^= *(uint32_t*)(((uintptr_t*)from) + offset);
#ifdef PERF_COUNT_XOR
		m_nbXor++;
#endif
	} 

#else //defined (__LP64__) || (__WORDSIZE == 64)

	// 32-bit machines
	for (offset = 0; offset < m_symbolSize32; offset++) {
		*(((uintptr_t*)to) + offset) ^=  *(((uintptr_t*)from) + offset);
#ifdef PERF_COUNT_XOR
		m_nbXor++;
#endif
	}
#endif //defined (__LP64__) || (__WORDSIZE == 64)
}


/******************************************************************************
 * BuildParitySymbol: Builds a new parity symbol.
 * => See header file for more informations.
 */
ldpc_error_status
LDPCFecSession::BuildParitySymbol (void* symbol_canvas[],
				int paritySymbol_index,
				void* paritySymbol)
{
	uintptr_t	*fec_buf;	// buffer for this parity symbol
	uintptr_t	*to_add_buf;	// buffer for the  source.parity symbol to add
        mod2entry	*e;
	int seqno;

	ASSERT(paritySymbol_index >= 0);
	ASSERT(paritySymbol_index < m_nbParitySymbols);
	ASSERT(paritySymbol != NULL);
	ASSERT(m_initialized);
	ASSERT(m_sessionFlags & FLAG_CODER);

	fec_buf = (uintptr_t*)GetBufferPtrOnly(paritySymbol);
	memset(fec_buf, 0, m_symbolSize);	// reset buffer (security)

	ASSERT(m_sessionType == TypeSTAIRS ||
		m_sessionType == TypeTRIANGLE ||
		m_sessionType == TypeLDGM);
	e = mod2sparse_first_in_row(m_pchkMatrix, paritySymbol_index);
	ASSERT(!mod2sparse_at_end(e));
	while (!mod2sparse_at_end(e)) {
		// paritySymbol_index in {0.. n-k-1} range, so this test is ok
		if (e->col != paritySymbol_index) {
			// don't add paritySymbol to itself
			seqno = GetSymbolSeqno(e->col);
			to_add_buf = (uintptr_t*)
					GetBuffer(symbol_canvas[seqno]);
			if (to_add_buf == NULL) {
				fprintf(stderr, "LDPCFecSession::BuildParitySymbol: FATAL ERROR, symbol %d is not allocated!\n", seqno);
				return LDPC_ERROR;
			}
			AddToSymbol(fec_buf, to_add_buf);
		}
		e = mod2sparse_next_in_row(e);
	}
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
	if (m_storeData_callback) {
		m_storeData_callback(m_context_4_callback, paritySymbol);
	}
#endif
#ifdef DEBUG
	if (this->m_verbosity >= 1) {
		printf("LDPCFecSession::BuildParitySymbol: parity symbol seq=%d created\n",
			paritySymbol_index);
	}
#endif
	return LDPC_OK;
}


/******************************************************************************
 * BuildParitySymbol: Builds a new parity symbol.
 * => See header file for more informations.
 */
ldpc_error_status
LDPCFecSession::BuildParitySymbolsPerCol (void*	symbol_canvas[],
					int	symbol_index,
					int*	built_parity_symbols[],
					int*	nb_built_parity_symbols)
{

	mod2entry	*e;
	uintptr_t	*data;
	uintptr_t	*paritySymbol;

	ASSERT(m_initialized);
	ASSERT(m_sessionFlags & FLAG_CODER);

	ASSERT(m_sessionType== TypeSTAIRS || m_sessionType == TypeTRIANGLE || m_sessionType== TypeLDGM);

	*nb_built_parity_symbols = 0;	
#ifdef DEBUG
	if (this->m_verbosity >= 1) {
		printf("LDPCFecSession::BuildParitySymbolsPerCol: column=%d processed\n", symbol_index);
	}
#endif
	e = mod2sparse_first_in_col(m_pchkMatrix, GetMatrixCol(symbol_index));
	ASSERT(!mod2sparse_at_end(e));

	while (!mod2sparse_at_end(e)) {
		if (e->row != GetMatrixCol(symbol_index)) {
			data = (uintptr_t*) GetBuffer(symbol_canvas[symbol_index]);
			if (symbol_canvas[GetSymbolSeqno(e->row)] == NULL) {
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				if (m_allocTmpBuffer_callback)
					symbol_canvas[GetSymbolSeqno(e->row)] = m_allocTmpBuffer_callback(m_context_4_callback,m_symbolSize);
				else
#endif
					symbol_canvas[GetSymbolSeqno(e->row)] = (char*)calloc(m_symbolSize, 1);
			}
			paritySymbol = (uintptr_t*) GetBufferPtrOnly(symbol_canvas[GetSymbolSeqno(e->row)]);
		
			if (data == NULL) {
				fprintf(stderr, "LDPCFecSession::BuildParitySymbol: FATAL ERROR, symbol %d is not allocated!\n", symbol_index);
				return LDPC_ERROR;
			}
			AddToSymbol(paritySymbol, data);
			m_nb_unknown_symbols_encoder[e->row]--;
			if (m_nb_unknown_symbols_encoder[e->row] == 1) {
				(*nb_built_parity_symbols)++;
				if (*nb_built_parity_symbols == 1) {
					*built_parity_symbols = (int*)
						calloc(1,sizeof(int));
				} else {
					*built_parity_symbols = (int*)
						realloc((void*) *built_parity_symbols, (*nb_built_parity_symbols)*sizeof(int));
				}
				(*built_parity_symbols)[(*nb_built_parity_symbols) - 1] = e->row;
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				if (m_storeData_callback) {
					m_storeData_callback(m_context_4_callback, paritySymbol);
				}
#endif						
			}
			
		}		
		e = mod2sparse_next_in_col(e);
	}
	
	/* Call recursively this function with new parity symbols as parameter*/
	int temp = *nb_built_parity_symbols;
	for (int i = 0; i < temp; i++) {
		int	recursive_nb_built_parity_symbols = 0;
		int	*recursive_built_parity_symbols = NULL;

		BuildParitySymbolsPerCol(symbol_canvas,
					*built_parity_symbols[i] + m_nbSourceSymbols,
					&recursive_built_parity_symbols,
					&recursive_nb_built_parity_symbols);
		*built_parity_symbols = (int*) realloc((void*) *built_parity_symbols,((*nb_built_parity_symbols) + recursive_nb_built_parity_symbols)*sizeof(int));
		for (int j = 0; j < recursive_nb_built_parity_symbols; j++) {
			(*built_parity_symbols)[*nb_built_parity_symbols+j] =
						recursive_built_parity_symbols[j];
		}
		*nb_built_parity_symbols = *nb_built_parity_symbols +
					recursive_nb_built_parity_symbols;

		if (recursive_built_parity_symbols != NULL) {
			free(recursive_built_parity_symbols);
			recursive_built_parity_symbols = NULL;
		}
	}
	return LDPC_OK;
}


/******************************************************************************
 * GetMatrixCol:
 * => See header file for more informations.
 */
int
LDPCFecSession::GetMatrixCol(int symbolSeqno)
{
	if (symbolSeqno < m_nbSourceSymbols) {
		/* source symbol */
		return (symbolSeqno + m_nbParitySymbols);
	} else {
		/* parity symbol */
		return (symbolSeqno - m_nbSourceSymbols);
	}
}


/******************************************************************************
 * GetSymbolSeqno:
 * => See header file for more informations.
 */
int
LDPCFecSession::GetSymbolSeqno(int matrixCol)
{
	int colInOrder;

	colInOrder = matrixCol;
	if (colInOrder < m_nbParitySymbols) {
		/* parity symbol */
		return (colInOrder + m_nbSourceSymbols);
	} else {
		/* source symbol */
		return (colInOrder - m_nbParitySymbols);
	}
}


/**
 * DecodingStepWithSymbol: Perform a new decoding step with a new (given) symbol.
 * This is the legacy front end to the DecodingStepWithSymbol() method. The actual
 * work will be done in the other DecodingStepWithSymbol() method.
 * => See header file for more informations.
 */ 
ldpc_error_status
LDPCFecSession::DecodingStepWithSymbol (void*	symbol_canvas[],
					void*	new_symbol,
					int	new_symbol_seqno,
					bool	store_symbol)
{
	void	*new_symbol_dst;	// temp variable used to store symbol

	ASSERT(new_symbol);
	ASSERT(new_symbol_seqno >= 0);
	ASSERT(new_symbol_seqno < m_nbTotalSymbols);
	ASSERT(m_initialized);
	ASSERT(m_sessionFlags & FLAG_DECODER);

	// Fast path. If store symbol is not set, then call directly
	// the full DecodingStepWithSymbol() method to avoid duplicate processing.
	if (store_symbol == false) {
		return(DecodingStepWithSymbol(symbol_canvas, new_symbol, new_symbol_seqno)); 
	}
	// Step 0: check if this is a fresh symbol, otherwise return
	if ((mod2sparse_last_in_col(m_pchkMatrix, GetMatrixCol(new_symbol_seqno))->row < 0)
	    || (IsSourceSymbol(new_symbol_seqno) && (symbol_canvas[new_symbol_seqno] != NULL))
	    || (IsParitySymbol(new_symbol_seqno) && (m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols] != NULL))) {
		// Symbol has already been processed, so skip it
#ifdef DEBUG
		if (this->m_verbosity >= 1) {
			printf("LDPCFecSession::DecodingStepWithSymbol: %s symbol %d already received or rebuilt, ignored\n",
				(IsSourceSymbol(new_symbol_seqno)) ? "Source" : "Parity",
				new_symbol_seqno); }
#endif
		return LDPC_OK;
	}
	// Step 1: Store the symbol in a permanent array if the caller wants it.
	// It concerns only source symbols, since parity symbols are only stored in
	// permanent array if we have a memory gain by doing so, which will
	// be defined later on in the full DecodingStepWithSymbol() method.
	if (IsSourceSymbol(new_symbol_seqno)) {
		ASSERT(store_symbol);
		// Call any required callback, or allocate memory, and
		// copy the symbol content in it.
		// This is typically something which is done when this
		// function is called recursively, for newly decoded
		// symbols.
		if (this->m_decodedSymbol_callback != NULL) {
			new_symbol_dst = m_decodedSymbol_callback(
						m_context_4_callback,
						m_symbolSize,
						new_symbol_seqno);
		} else {
			new_symbol_dst = (void *)malloc(m_symbolSize);
		}
		if (new_symbol_dst == NULL) {
			fprintf(stderr, "LDPCFecSession::DecodingStepWithSymbol: ERROR, out of memory!\n");
			return LDPC_ERROR;
		}
		// Copy data now
		memcpy(GetBufferPtrOnly(new_symbol_dst), GetBuffer(new_symbol), m_symbolSize);
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
		if (m_storeData_callback != NULL) {
			m_storeData_callback(m_context_4_callback,
					new_symbol_dst);
		}
#endif
	} else {
		new_symbol_dst = new_symbol;
	}
	/* continue decoding with the full DecodingStepWithSymbol() method */
	return(DecodingStepWithSymbol(symbol_canvas, new_symbol_dst, new_symbol_seqno)); 
}


/******************************************************************************
 * DecodingStepWithSymbol: Perform a new decoding step with a new (given) symbol.
 * => See header file for more informations.
 *
 * This function relies on the following simple algorithm:
 *
 * Given a set of linear equations, if one of them has only one
 * remaining unknown variable, then the value of this variable is
 * that of the constant term.
 * Replace this variable by its value in all remaining linear
 * equations, and reiterate. The value of several variables can
 * therefore be found by this recursive algorithm.
 *
 * In practice, an incoming symbol contains the value of the associated
 * variable, so replace its value in all linear equations in which
 * it is implicated. Then apply the above algorithm and see if decoding
 * can progress by one or more steps.
 *
 * For instance, if {s1, s2} are source symbols, and {f1} a parity symbol:
 *    { s1 + s2 + f1 = A      (eq. 1)
 *    {      s2 + f1 = B      (eq. 2)
 * Now if the node receives symbol s2, he replaces its value in the two
 * equations, he then finds f1, he replaces its value in the first equation
 * and finds s1.
 */
ldpc_error_status
LDPCFecSession::DecodingStepWithSymbol(	void*	symbol_canvas[],
				void*	new_symbol,
				int	new_symbol_seqno)
{
	mod2entry	*e = NULL;	// entry ("1") in parity check matrix
	mod2entry	*delMe;		// temp: entry to delete in row/column
	void		*currChk;	// temp: pointer to Partial sum
	int		row;		// temp: current row value
	int		*CheckOfDeg1 = NULL; // table of check nodes of degree
					// one after the processing of new_symbol
	int		CheckOfDeg1_nb = 0; // number of entries in table
	int		CheckOfDeg1_listSize = 0; // size of the memory block
					// allocated for the table
#ifdef PART_SUM_OPTIMIZATION
	bool		keep_symbol;	// true if it's worth to store new_symbol
					// in this function, in case it's a parity
					// symbol, and independantly from the
					// store_symbol argument.
#endif

	ASSERT(new_symbol);
	ASSERT(new_symbol_seqno >= 0);
	ASSERT(new_symbol_seqno < m_nbTotalSymbols);
	ASSERT(m_initialized);
	ASSERT(m_sessionFlags & FLAG_DECODER);

	// Step 0: check if this is a fresh symbol, otherwise return
	if ((mod2sparse_last_in_col(m_pchkMatrix, GetMatrixCol(new_symbol_seqno))->row < 0)
	    || (IsSourceSymbol(new_symbol_seqno) && (symbol_canvas[new_symbol_seqno] != NULL))
	    || (IsParitySymbol(new_symbol_seqno) && (m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols] != NULL))) {
		// Symbol has already been processed, so skip it
#ifdef DEBUG
	  if (this->m_verbosity >= 1) {
		printf("LDPCFecSession::DecodingStepWithSymbol: %s symbol %d already received or rebuilt, ignored\n",
		  				(IsSourceSymbol(new_symbol_seqno)) ? "Source" : "Parity",
		  				new_symbol_seqno);
	  }
#endif
		return LDPC_OK;
	}
#ifdef DEBUG
	if (this->m_verbosity >= 1) {
		printf("LDPCFecSession::DecodingStepWithSymbol: Processing NEW %s symbol: seq=%d\n",
			(IsSourceSymbol(new_symbol_seqno)) ? "Source" : "Parity",
			new_symbol_seqno);
	}
#endif
	// First, make sure data is available for this new symbol. Must
	// remain valid throughout this function...
	GetBuffer(new_symbol);

	// Step 1: Store the symbol in a permanent array. It concerns only DATA
	// symbols. Parity symbols are only stored in permanent array, if we have
	// a memory gain by doing so (and not creating new partial sums)
	if (IsSourceSymbol(new_symbol_seqno)) {
		// Source symbol
		// There's no need to allocate anything, nor to call
		// anything. It has already been done by the caller...
		symbol_canvas[new_symbol_seqno] = new_symbol;
#ifdef PART_SUM_OPTIMIZATION
		keep_symbol = true;
#endif
	} else {
		// Parity symbol
		// Check if parity symbol should be stored or if partial
		// sum should be stored
#ifdef PART_SUM_OPTIMIZATION
		if (m_triangleWithSmallFECRatio) {
			// In this case, the symbol will never be stored into
			// permanent array, but directly added to partial sum
			keep_symbol = false;
		} else {
			// The symbol will be stored if more than 1 partial sum
			// is needed
			const int	max_allowed_PS = 1; // threshold before 
							// storing symbol in array
			int		PS_to_create = 0; // nb of partial sums that
							//  would have to be allocated

			// count the number of PS that would be created
			// if we don't keep this parity symbol.
			for (e = mod2sparse_first_in_col(m_pchkMatrix,
							GetMatrixCol(new_symbol_seqno));
			    !mod2sparse_at_end(e);
			    e = mod2sparse_next_in_col(e))
			{
				if (m_checkValues[e->row] == NULL &&
				    m_nb_unknown_symbols[e->row] > 2) {
					PS_to_create++;
					if (PS_to_create > max_allowed_PS) {
						break;	// no need to see further
					}
				}
			}
			// now take a decision...
			if (PS_to_create > max_allowed_PS) {
				// Parity symbol will be stored in a permanent array
				// Alloc the buffer...
				keep_symbol = true;
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				if (m_allocTmpBuffer_callback != NULL) {
					m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols] =
							m_allocTmpBuffer_callback(
								m_context_4_callback,
								m_symbolSize);
				} else
#endif //EXTERNAL_MEMORY_MGMT_SUPPORT
				{
					m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols] =
							(void *)malloc(m_symbolSize);
				}
				// copy the content...
				memcpy(GetBufferPtrOnly(m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols]),
					GetBufferPtrOnly(new_symbol), m_symbolSize);
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				// and store it permanently.
				if (m_storeData_callback) {
					m_storeData_callback(m_context_4_callback,
							m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols]);
				}
#endif //EXTERNAL_MEMORY_MGMT_SUPPORT
			} else {
				// Parity symbol will only be added to partial sums
				keep_symbol = false;
			}
		}
#else  // !PART_SUM_OPTIMIZATION
		if (m_triangleWithSmallFECRatio) {
			// In this case, the symbol will never be stored into
			// permanent array, but directly added to partial sum
		} else {
			// Check if parity symbol should be stored or if
			// partial sum should be stored
			bool	store_parity = false;
			for (e = mod2sparse_first_in_col(m_pchkMatrix,
					    GetMatrixCol(new_symbol_seqno));
			    !mod2sparse_at_end(e);
			    e = mod2sparse_next_in_col(e))
			{
				if (m_nb_unknown_symbols[e->row] > 2) {
					store_parity = true;
					break;
				}
			}
			if (store_parity) {
				// Parity symbol will be stored in a permanent array
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				if (m_allocTmpBuffer_callback != NULL) {
					m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols] =
							m_allocTmpBuffer_callback(
								m_context_4_callback,
								m_symbolSize);
				} else
#endif //EXTERNAL_MEMORY_MGMT_SUPPORT
				{
					m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols] =
							(void *)malloc(m_symbolSize);
				}
				// copy the content...
				memcpy(GetBufferPtrOnly(m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols]),
					GetBufferPtrOnly(new_symbol), m_symbolSize);
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				// and store it permanently.
				if (m_storeData_callback) {
					m_storeData_callback(m_context_4_callback,
							m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols]);
				}
#endif //EXTERNAL_MEMORY_MGMT_SUPPORT
			}
			// else parity symbol will only be added to partial sums
		}
#endif  // !PART_SUM_OPTIMIZATION
	}

	// Step 2: Inject the symbol value in each equation it is involved
	// (if partial sum already exists or if partial sum should be created)
	for (e = mod2sparse_first_in_col(m_pchkMatrix, GetMatrixCol(new_symbol_seqno));
	     !mod2sparse_at_end(e); ) {
		// for a given row, ie for a given equation where this symbol
		// is implicated, do the following:
		row = e->row;
		m_nb_unknown_symbols[row]--;	// symbol is known
		currChk = m_checkValues[row];	// associated check
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
		if (currChk != NULL) {
			// make sure data is available
			if (m_getData_callback != NULL) {
				m_getData_callback(m_context_4_callback, currChk);
			}
		}
#endif
		if (currChk == NULL &&
#ifdef PART_SUM_OPTIMIZATION
		    ((keep_symbol == false) || (m_nb_unknown_symbols[row] == 1))
#else
		    ((m_nb_unknown_symbols[row] == 1) || m_triangleWithSmallFECRatio)
#endif
		) {
			// we need to allocate a PS (i.e. check node)
			// and add symbol to it, because the parity symbol
			// won't be kept (keep_symbol == false), or it is the
			// last missing symbol of this equation, or because
			// or some particular situation where it is non sense
			// no to allocate a PS (m_triangleWithSmallFECRatio).
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
			if (m_allocTmpBuffer_callback != NULL) {
				currChk = m_allocTmpBuffer_callback(
							m_context_4_callback,
							m_symbolSize);
			} else
#endif
			{
				currChk = (void*) calloc(m_symbolSize32, 4);
			}
			if ((m_checkValues[row] = currChk) == NULL) {
				goto no_mem;
			}
		}
		if (currChk != NULL) {
			// there's a partial sum for this row...
			if (m_nbSymbols_in_equ[row] > 1) {
				// we can add the symbol content to this PS
				AddToSymbol(GetBufferPtrOnly(currChk),
					    GetBufferPtrOnly(new_symbol));
			}
			// else this is useless, since new_symbol is the last
			// symbol of this equation, and its value is necessary
			// equal to the PS. Their sum must be 0 (we don't
			// check it).

			// remove the symbol from the equation (this entry
			// is now useless)
			delMe = e;
			e = mod2sparse_next_in_col(e);
			mod2sparse_delete(m_pchkMatrix, delMe);
			m_nbSymbols_in_equ[row]--;
			if (IsParitySymbol(new_symbol_seqno)) {
				m_nbEqu_for_parity[new_symbol_seqno - m_nbSourceSymbols]--;
			}

			// Inject all permanently stored symbols (DATA and parity)
			// into partial sum
#ifdef PART_SUM_OPTIMIZATION
			if ((m_nb_unknown_symbols[row] == 1) ||
			    ((keep_symbol == false) && (m_triangleWithSmallFECRatio == false)))
#else
			if (m_triangleWithSmallFECRatio == false)
#endif
			{
				// Inject all permanently stored symbols
				// (DATA and parity) into this partial sum.
				// Requires to scan the equation (ie row).
				mod2entry	*tmp_e;	// curr symbol in this equ
				int		tmp_seqno;// corresponding seq no
				void		*tmp_symbol; // corresponding symbol pointer

				for (tmp_e = mod2sparse_first_in_row(m_pchkMatrix, row);
				     !mod2sparse_at_end(tmp_e); ) {

					tmp_seqno = GetSymbolSeqno(tmp_e->col);
					//if (GetSymbolSeqno(r->col) >= m_nbSourceSymbols)
					if (IsParitySymbol(tmp_seqno)) {
						tmp_symbol = m_parity_symbol_canvas[tmp_seqno - m_nbSourceSymbols];
					} else {
						// waiting for
						// (m_nb_unknown_symbols[row] == 1)
						// to add source symbols is
						// useless... it's even slower
						// in that case!
						tmp_symbol = symbol_canvas[tmp_seqno];
					}
					if (tmp_symbol != NULL) {
						// add the symbol content now
						AddToSymbol(
							GetBufferPtrOnly(currChk),
							GetBuffer(tmp_symbol));
						// delete the entry
						delMe = tmp_e;
						tmp_e =  mod2sparse_next_in_row(tmp_e);
						mod2sparse_delete(m_pchkMatrix, delMe);
						m_nbSymbols_in_equ[row]--;
						if (IsParitySymbol(tmp_seqno)) {
							m_nbEqu_for_parity[tmp_seqno - m_nbSourceSymbols]--;
							// check if we can delete
							// parity symbol altogether
							if (m_nbEqu_for_parity[tmp_seqno - m_nbSourceSymbols] == 0) {
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
								if (m_freeSymbol_callback != NULL) {
									m_freeSymbol_callback(
										m_context_4_callback,
										tmp_symbol);
								} else
#endif
								{
									free(tmp_symbol);
								}
								m_parity_symbol_canvas[tmp_seqno - m_nbSourceSymbols] = NULL;
							}
						}
					} else {
						// this symbol not yet known,
						// switch to next one in equ
						tmp_e =  mod2sparse_next_in_row(tmp_e);
					}
				}
			}
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
			// store the partial sum now we are sure it has been
			// completely updated
			if (m_storeData_callback != NULL) {
				m_storeData_callback(m_context_4_callback,
							currChk);
			}
#endif
		} else {
			// here m_checkValues[row] is NULL, ie. the partial
			// sum has not been allocated
			e = mod2sparse_next_in_col(e);
		}
		if (m_nbSymbols_in_equ[row] == 1) {
			// register this entry for step 3 since the symbol
			// associated to this equation can now be decoded...
			if (CheckOfDeg1 == NULL) {
				// allocate memory for the table first
				CheckOfDeg1_listSize = 4;
				if ((CheckOfDeg1 = (int*)
						calloc(CheckOfDeg1_listSize,
						sizeof(int*))) == NULL) {
					goto no_mem;
				}
			} else if (CheckOfDeg1_nb == CheckOfDeg1_listSize) {
				// not enough size in table, add some more
				CheckOfDeg1_listSize += 4;
				if ((CheckOfDeg1 = (int*)realloc(CheckOfDeg1,
							CheckOfDeg1_listSize * sizeof(int*))) == NULL) {
					goto no_mem;
				}
			}
			CheckOfDeg1[CheckOfDeg1_nb++] = row;
		}
	}

	// Step 3: Check if a new symbol has been decoded and take appropriate
	// measures ...
	int	decoded_symbol_seqno;	// sequence number of decoded symbol
	//for (int i = 0; i < CheckOfDeg1_nb; i++) 
	for (CheckOfDeg1_nb--; CheckOfDeg1_nb >= 0; CheckOfDeg1_nb--) {
		// get the index (ie row) of the partial sum concerned
		row = CheckOfDeg1[CheckOfDeg1_nb];
		if (m_nbSymbols_in_equ[row] == 1) {
			// A new decoded symbol is available...
			// NB: because of the recursion below, we need to
			// check that all equations mentioned in the
			// CheckOfDeg1 list are __still__ of degree 1.
			e = mod2sparse_first_in_row(m_pchkMatrix, row);
			ASSERT(!mod2sparse_at_end(e) &&
				mod2sparse_at_end(e->right))
			decoded_symbol_seqno = GetSymbolSeqno(e->col);
			// remove the entry from the matrix
			currChk = m_checkValues[row];	// remember it
			m_checkValues[row] = NULL;
			m_nbSymbols_in_equ[row]--;
			if (IsParitySymbol(decoded_symbol_seqno)) {
				m_nbEqu_for_parity[decoded_symbol_seqno - m_nbSourceSymbols]--;
			}
			mod2sparse_delete(m_pchkMatrix, e);
#ifdef DEBUG
			if (this->m_verbosity >= 1) {
				printf("LDPCFecSession::DecodingStepWithSymbol: => REBUILT %s symbol %d\n",
					(IsParitySymbol(decoded_symbol_seqno)) ? "Parity" : "Source",
					decoded_symbol_seqno);
			}
#endif
			if (IsSourceSymbol(decoded_symbol_seqno)) {
				// source symbol.
				void	*decoded_symbol_dst;// temp variable used to store symbol

				// First copy it into a permanent symbol.
				// Call any required callback, or allocate memory, and
				// copy the symbol content in it.
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				if (this->m_decodedSymbol_callback != NULL) {
					decoded_symbol_dst =
						m_decodedSymbol_callback(
							m_context_4_callback,
							m_symbolSize,
							decoded_symbol_seqno);
				} else
#endif
				{
					decoded_symbol_dst =
						(void *)malloc(m_symbolSize);
				}
				if (decoded_symbol_dst == NULL) {
					goto no_mem;
				}
				memcpy(GetBufferPtrOnly(decoded_symbol_dst),
					GetBuffer(currChk), m_symbolSize);
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				if (m_storeData_callback != NULL) {
					m_storeData_callback(m_context_4_callback,
							decoded_symbol_dst);
				}
#endif
				// Free partial sum which is no longer used.
				// It's important to free it before calling
				// DecodingStepWithSymbol recursively to reduce max
				// memory requirements.
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				if (m_freeSymbol_callback != NULL) {
					m_freeSymbol_callback(m_context_4_callback,
							currChk);
				} else
#endif
				{
					free(currChk);	
				}
				// And finally call this method recursively if
				// decoding is not completed... Otherwise, it's
				// useless, the source block is now decoded.
				if(!IsDecodingComplete(symbol_canvas)) {
					DecodingStepWithSymbol(symbol_canvas, decoded_symbol_dst,
								decoded_symbol_seqno);
				}

			} else {

				// Parity symbol.
				// Call this method recursively first if
				// decoding is not completed... Otherwise, it's
				// useless, the source block is now decoded.
				// Call this method recursively first...
				if(!IsDecodingComplete(symbol_canvas)) {
					DecodingStepWithSymbol(symbol_canvas, currChk,
								decoded_symbol_seqno);
				}
				// Then free the partial sum which is no longer needed.
#ifdef EXTERNAL_MEMORY_MGMT_SUPPORT
				if (m_freeSymbol_callback != NULL) {
					m_freeSymbol_callback(m_context_4_callback,
							currChk);
				} else
#endif
				{
					free(currChk);	
				}
			}
		}
	}
	if (CheckOfDeg1 != NULL) {
		free(CheckOfDeg1);
	}
	return LDPC_OK;

no_mem:
	fprintf(stderr, "LDPCFecSession::DecodingStepWithSymbol: ERROR, out of memory!\n");
	return LDPC_ERROR;
}


/******************************************************************************
 * IsAlreadyKnown: Returns true if the symbol has already been received
 * or decoded (i.e. if it is already known), false otherwise.
 * => See header file for more informations.
 */
bool
LDPCFecSession::SymbolAlreadyKnown (void*	symbol_canvas[],
				    int		new_symbol_seqno)
{
	if ((mod2sparse_last_in_col(m_pchkMatrix, GetMatrixCol(new_symbol_seqno))->row < 0)
	    || (IsSourceSymbol(new_symbol_seqno) && (symbol_canvas[new_symbol_seqno] != NULL))
	    || (IsParitySymbol(new_symbol_seqno) && (m_parity_symbol_canvas[new_symbol_seqno - m_nbSourceSymbols] != NULL))) {
		// No entry in the column associated to this symbol.
		// Means symbol has already been processed, so skip it.
#ifdef DEBUG
		if (this->m_verbosity >= 1) {
			printf("LDPCFecSession::SymbolAlreadyKnown: %s symbol %d already received or rebuilt\n",
				(new_symbol_seqno < m_nbSourceSymbols) ? "Source" : "Parity",
				new_symbol_seqno);
		}
#endif
		return true;
	} else {
#ifdef DEBUG
		if (this->m_verbosity >= 1) {
			printf("LDPCFecSession::SymbolAlreadyKnown: %s symbol %d not received or rebuilt\n",
				(new_symbol_seqno < m_nbSourceSymbols) ? "Source" : "Parity",
				new_symbol_seqno);
		}
#endif
		return false;
	}
}


/******************************************************************************
 * IsDecodingComplete: Checks if all DATA symbols have been received/rebuilt.
 * => See header file for more informations.
 */
bool
LDPCFecSession::IsDecodingComplete(void*	symbol_canvas[])
{
	if (!m_initialized) {
		fprintf(stderr, "LDPCFecSession::IsDecodingComplete: ERROR: LDPC Session is NOT initialized!\n");
		return false;
	}

	for (int i = m_firstNonDecoded; i < m_nbSourceSymbols; i++) {
		if (symbol_canvas[i] == NULL) {
			/* not yet decoded! */
			m_firstNonDecoded = i;	/* remember for next time */
			return false;
		}
	}
	return true;
}

